<?php
/**
 * Custom Post Types field for ACF.
 * https://github.com/TimPerry/acf-post-type-selector
 *
 * @package WordPress
 * @version 1.0
 */

/**
 * Add a selectable CPT field to pages.
 */
class ACF_CPT extends acf_field {
	const SELECTOR_TYPE_SELECT = 0;
	/**
	 * __construct
	 *
	 * This function will setup the field type data
	 *
	 * @type function
	 * @date 5/03/2014
	 * @since 5.0.0
	 *
	 * @return n/a
	 */
	public function __construct() {
		// vars.
		$this->name     = 'post_type_selector';
		$this->label    = __( 'Post Type Selector' );
		$this->category = __( 'Relational', 'acf' ); // Basic, Content, Choice, etc.
		$this->defaults = array(
			'select_type' => self::SELECTOR_TYPE_SELECT,
		);
		// Do not delete!
		parent::__construct();
	}
	/**
	 * Start render_field()
	 * Create the HTML interface for your field.
	 *
	 * @param $field (array) the $field being rendered.
	 *
	 * @type action
	 * @since 3.6
	 * @date 23/01/13
	 */
	public function render_field( $field ) {
		$field      = array_merge( $this->defaults, $field ); // Defaults?
		$post_types = acf_get_pretty_post_types();
		/**
		 * Filters the array of post types.
		 *
		 * @since 1.0.1
		 *
		 * @param array $post_types List of post types.
		 * @param array $field      The field being rendered.
		 */
		$post_types = apply_filters( 'post_type_selector_post_types', $post_types, $field );
		// Not required: add empty/none value.
		if ( ! $field['required'] ) {
			$post_types = [ '' => __( 'None', 'acf' ) ] + $post_types;
		}
		// Create Field HTML.
		$checked = array();
		echo '<select id="' . esc_html( $field['name'] ) . '" class="' . esc_html( $field['class'] ) . '" name="' . esc_html( $field['name'] ) . '">';
		$checked[ $field['value'] ] = 'selected="selected"';
		foreach ( $post_types as $post_type => $post_type_label ) {
			echo '<option ' . ( isset( $checked[ $post_type ] ) ? $checked[ $post_type ] : null ) . ' value="' . esc_html( $post_type ) . '">' . esc_html( $post_type_label ) . '</option>'; // phpcs:ignore
		}
		echo '</select>';
	}
	/**
	 * Start load_value()
	 *
	 * This filter is applied to the $value after it is loaded from the db
	 *
	 * @type filter
	 * @since 3.6
	 * @date 23/01/13
	 *
	 * @param array $value (mixed) the value found in the database.
	 * @param array $post_id (mixed) the $post_id from which the value was loaded.
	 * @param array $field (array) the field array holding all the field options.
	 * @return $value
	 */
	public function load_value( $value, $post_id, $field ) {
		return $value;
	}
	/**
	 * Start update_value()
	 *
	 * This filter is applied to the $value before it is saved in the db
	 *
	 * @type filter
	 * @since 3.6
	 * @date 23/01/13
	 *
	 * @param array $value (mixed) the value found in the database.
	 * @param array $post_id (mixed) the $post_id from which the value was loaded.
	 * @param array $field (array) the field array holding all the field options.
	 * @return $value
	 */
	public function update_value( $value, $post_id, $field ) {
		return $value;
	}
	/**
	 * Start format_value()
	 *
	 * This filter is applied to the $value after it is loaded from the db
	 * and before it is returned to the template
	 *
	 * @type filter
	 * @since 3.6
	 * @date 23/01/13
	 *
	 * @param array $value (mixed) the value which was loaded from the database.
	 * @param array $post_id (mixed) the $post_id from which the value was loaded.
	 * @param array $field (array) the field array holding all the field options.
	 *
	 * @return $value (mixed) the modified value
	 */
	public function format_value( $value, $post_id, $field ) {
		return $value;
	}
	/**
	 * Start format_value_for_api()
	 *
	 * This filter is applied to the $value after it is loaded from the db
	 * and before it is passed back to the api functions such as the_field
	 *
	 * @type filter
	 * @since 3.6
	 * @date 23/01/13
	 *
	 * @param array $value - the value which was loaded from the database.
	 * @param array $post_id - the $post_id from which the value was loaded.
	 * @param array $field - the field array holding all the field options.
	 *
	 * @return $value - the modified value.
	 */
	public function format_value_for_api( $value, $post_id, $field ) {
		return $value;
	}
	/**
	 * Start load_field()
	 * This filter is applied to the $field after it is loaded from the database
	 *
	 * @type filter
	 * @date 23/01/2013
	 * @since 3.6.0
	 *
	 * @param array $field (array) the field array holding all the field options.
	 * @return $field
	 */
	public function load_field( $field ) {
		return $field;
	}
	/**
	 * Start update_field()
	 * This filter is applied to the $field before it is saved to the database
	 *
	 * @type filter
	 * @date 23/01/2013
	 * @since 3.6.0
	 *
	 * @param array $field (array) the field array holding all the field options.
	 * @return $field
	 */
	public function update_field( $field ) {
		return $field;
	}
}
// Create field.
new ACF_CPT();

